import logging
import time
from collections import defaultdict
from queue import Queue
from threading import Thread

from kube_hunter.conf import get_config
from kube_hunter.core.types import ActiveHunter, HunterBase
from kube_hunter.core.events.types import Vulnerability, EventFilterBase, MultipleEventsContainer

logger = logging.getLogger(__name__)


# Inherits Queue object, handles events asynchronously
class EventQueue(Queue):
    def __init__(self, num_worker=10):
        super().__init__()
        self.passive_hunters = dict()
        self.active_hunters = dict()
        self.all_hunters = dict()

        self.running = True
        self.workers = list()

        # -- Regular Subscription --
        # Structure: key: Event Class, value: tuple(Registered Hunter, Predicate Function)
        self.hooks = defaultdict(list)
        self.filters = defaultdict(list)
        # --------------------------

        # -- Multiple Subscription --
        # Structure: key: Event Class, value: tuple(Registered Hunter, Predicate Function)
        self.multi_hooks = defaultdict(list)

        # When subscribing to multiple events, this gets populated with required event classes
        # Structure: key: Hunter Class, value: set(RequiredEventClass1, RequiredEventClass2)
        self.hook_dependencies = defaultdict(set)

        # To keep track of fulfilled dependencies. we need to have a structure which saves historical instanciated
        # events mapped to a registered hunter.
        # We used a 2 dimensional dictionary in order to fulfill two demands:
        #   * correctly count published required events
        #   * save historical events fired, easily sorted by their type
        #
        # Structure: hook_fulfilled_deps[hunter_class] -> fulfilled_events_for_hunter[event_class] -> [EventObject, EventObject2]
        self.hook_fulfilled_deps = defaultdict(lambda: defaultdict(list))
        # ---------------------------

        for _ in range(num_worker):
            t = Thread(target=self.worker)
            t.daemon = True
            t.start()
            self.workers.append(t)

        t = Thread(target=self.notifier)
        t.daemon = True
        t.start()

    """
    ######################################################
    + ----------------- Public Methods ----------------- +
    ######################################################
    """

    def subscribe(self, event, hook=None, predicate=None, is_register=True):
        """
        The Subscribe Decorator - For Regular Registration
        Use this to register for one event only. Your hunter will execute each time this event is published

        @param event - Event class to subscribe to
        @param predicate - Optional: Function that will be called with the published event as a parameter before trigger.
                            If it's return value is False, the Hunter will not run (default=None).
        @param hook - Hunter class to register for (ignore when using as a decorator)
        """

        def wrapper(hook):
            self.subscribe_event(event, hook=hook, predicate=predicate, is_register=is_register)
            return hook

        return wrapper

    def subscribe_many(self, events, hook=None, predicates=None, is_register=True):
        """
        The Subscribe Many Decorator - For Multiple Registration,
        When your attack needs several prerequisites to exist in the cluster, You need to register for multiple events.
        Your hunter will execute once for every new combination of required events.
        For example:
            1. event A was published 3 times
            2. event B was published once.
            3. event B was published again
        Your hunter will execute 2 times:
            * (on step 2) with the newest version of A
            * (on step 3) with the newest version of A and newest version of B

        @param events - List of event classes to subscribe to
        @param predicates - Optional: List of function that will be called with the published event as a parameter before trigger.
                            If it's return value is False, the Hunter will not run (default=None).
        @param hook - Hunter class to register for (ignore when using as a decorator)
        """

        def wrapper(hook):
            self.subscribe_events(events, hook=hook, predicates=predicates, is_register=is_register)
            return hook

        return wrapper

    def subscribe_once(self, event, hook=None, predicate=None, is_register=True):
        """
        The Subscribe Once Decorator - For Single Trigger Registration,
        Use this when you want your hunter to execute only in your entire program run
        wraps subscribe_event method

        @param events - List of event classes to subscribe to
        @param predicates - Optional: List of function that will be called with the published event as a parameter before trigger.
                            If it's return value is False, the Hunter will not run (default=None).
        @param hook - Hunter class to register for (ignore when using as a decorator)
        """

        def wrapper(hook):
            # installing a __new__ magic method on the hunter
            # which will remove the hunter from the list upon creation
            def __new__unsubscribe_self(self, cls):
                handler.hooks[event].remove((hook, predicate))
                return object.__new__(self)

            hook.__new__ = __new__unsubscribe_self

            self.subscribe_event(event, hook=hook, predicate=predicate, is_register=is_register)

            return hook

        return wrapper

    def publish_event(self, event, caller=None):
        """
        The Publish Event Method - For Publishing Events To Kube-Hunter's Queue
        """
        # Document that the hunter published a vulnerability (if it's indeed a vulnerability)
        # For statistics options
        self._increase_vuln_count(event, caller)

        # sets the event's parent to be it's publisher hunter.
        self._set_event_chain(event, caller)

        # applying filters on the event, before publishing it to subscribers.
        # if filter returned None, not proceeding to publish
        event = self.apply_filters(event)
        if event:
            # If event was rewritten, make sure it's linked again
            self._set_event_chain(event, caller)

            # Regular Hunter registrations - publish logic
            # Here we iterate over all the registered-to events:
            for hooked_event in self.hooks.keys():
                # We check if the event we want to publish is an inherited class of the current registered-to iterated event
                # Meaning - if this is a relevant event:
                if hooked_event in event.__class__.__mro__:
                    # If so, we want to publish to all registerd hunters.
                    for hook, predicate in self.hooks[hooked_event]:
                        if predicate and not predicate(event):
                            continue

                        self.put(hook(event))
                        logger.debug(f"Event {event.__class__} got published to hunter - {hook} with {event}")

            # Multiple Hunter registrations - publish logic
            # Here we iterate over all the registered-to events:
            for hooked_event in self.multi_hooks.keys():
                # We check if the event we want to publish is an inherited class of the current registered-to iterated event
                # Meaning - if this is a relevant event:
                if hooked_event in event.__class__.__mro__:
                    # now we iterate over the corresponding registered hunters.
                    for hook, predicate in self.multi_hooks[hooked_event]:
                        if predicate and not predicate(event):
                            continue

                        self._update_multi_hooks(hook, event)

                        if self._is_all_fulfilled_for_hunter(hook):
                            events_container = MultipleEventsContainer(self._get_latest_events_from_multi_hooks(hook))
                            self.put(hook(events_container))
                            logger.debug(
                                f"Multiple subscription requirements were met for hunter {hook}. events container was \
                                published with {self.hook_fulfilled_deps[hook].keys()}"
                            )

    """
    ######################################################
    + ---------------- Private Methods ----------------- +
    + ---------------- (Backend Logic) ----------------- +
    ######################################################
    """

    def _get_latest_events_from_multi_hooks(self, hook):
        """
        Iterates over fulfilled deps for the hunter, and fetching the latest appended events from history
        """
        latest_events = list()
        for event_class in self.hook_fulfilled_deps[hook].keys():
            latest_events.append(self.hook_fulfilled_deps[hook][event_class][-1])
        return latest_events

    def _update_multi_hooks(self, hook, event):
        """
        Updates published events in the multi hooks fulfilled store.
        """
        self.hook_fulfilled_deps[hook][event.__class__].append(event)

    def _is_all_fulfilled_for_hunter(self, hook):
        """
        Returns true for multi hook fulfilled, else oterwise
        """
        # Check if the first dimension already contains all necessary event classes
        return len(self.hook_fulfilled_deps[hook].keys()) == len(self.hook_dependencies[hook])

    def _set_event_chain(self, event, caller):
        """
        Sets' events attribute chain.
        In here we link the event with it's publisher (Hunter),
        so in the next hunter that catches this event, we could access the previous one's attributes.

        @param event: the event object to be chained
        @param caller: the Hunter object that published this event.
        """
        if caller:
            event.previous = caller.event
            event.hunter = caller.__class__

    def _register_hunters(self, hook=None):
        """
        This method is called when a Hunter registers itself to the handler.
        this is done in order to track and correctly configure the current run of the program.

        passive_hunters, active_hunters, all_hunters
        """
        config = get_config()
        if ActiveHunter in hook.__mro__:
            if not config.active:
                return False
            else:
                self.active_hunters[hook] = hook.__doc__
        elif HunterBase in hook.__mro__:
            self.passive_hunters[hook] = hook.__doc__

        if HunterBase in hook.__mro__:
            self.all_hunters[hook] = hook.__doc__

        return True

    def _register_filter(self, event, hook=None, predicate=None):
        if hook not in self.filters[event]:
            self.filters[event].append((hook, predicate))
            logging.debug("{} filter subscribed to {}".format(hook, event))

    def _register_hook(self, event, hook=None, predicate=None):
        if hook not in self.hooks[event]:
            self.hooks[event].append((hook, predicate))
            logging.debug("{} subscribed to {}".format(hook, event))

    def allowed_for_custom_registration(self, target_hunter):
        """
        Check if the partial input list contains the hunter we are about to register for events
        If hunter is considered a Core hunter as specified in `config.core_hunters` we allow it anyway

        Returns true if:
         1. partial hunt is disabled
         2. partial hunt is enabled and hunter is in core hunter class
         3. partial hunt is enabled and hunter is specified in config.partial

        @param target_hunter: hunter class for registration check
        """
        config = get_config()
        if not config.custom:
            return True

        hunter_class_name = target_hunter.__name__
        if hunter_class_name in config.core_hunters or hunter_class_name in config.custom:
            return True

        return False

    def subscribe_event(self, event, hook=None, predicate=None, is_register=True):
        if not is_register:
            return
        if not self.allowed_for_custom_registration(hook):
            return
        if not self._register_hunters(hook):
            return

        # registering filters
        if EventFilterBase in hook.__mro__:
            self._register_filter(event, hook, predicate)
        # registering hunters
        else:
            self._register_hook(event, hook, predicate)

    def subscribe_events(self, events, hook=None, predicates=None, is_register=True):
        if not is_register:
            return
        if not self.allowed_for_custom_registration(hook):
            return
        if not self._register_hunters(hook):
            return

        if predicates is None:
            predicates = [None] * len(events)

        # registering filters.
        if EventFilterBase in hook.__mro__:
            for event, predicate in zip(events, predicates):
                self._register_filter(event, hook, predicate)
        # registering hunters.
        else:
            for event, predicate in zip(events, predicates):
                self.multi_hooks[event].append((hook, predicate))

            self.hook_dependencies[hook] = frozenset(events)

    def apply_filters(self, event):
        # if filters are subscribed, apply them on the event
        for hooked_event in self.filters.keys():
            if hooked_event in event.__class__.__mro__:
                for filter_hook, predicate in self.filters[hooked_event]:
                    if predicate and not predicate(event):
                        continue

                    logger.debug(f"Event {event.__class__} filtered with {filter_hook}")
                    event = filter_hook(event).execute()
                    # if filter decided to remove event, returning None
                    if not event:
                        return None
        return event

    def _increase_vuln_count(self, event, caller):
        config = get_config()
        if config.statistics and caller:
            if Vulnerability in event.__class__.__mro__:
                caller.__class__.publishedVulnerabilities += 1

    # executes callbacks on dedicated thread as a daemon
    def worker(self):
        while self.running:
            try:
                hook = self.get()
                logger.debug(f"Executing {hook.__class__} with {hook.event.__dict__}")
                hook.execute()
            except Exception as ex:
                logger.debug(ex, exc_info=True)
            finally:
                self.task_done()
        logger.debug("closing thread...")

    def notifier(self):
        time.sleep(2)
        # should consider locking on unfinished_tasks
        while self.unfinished_tasks > 0:
            logger.debug(f"{self.unfinished_tasks} tasks left")
            time.sleep(3)
            if self.unfinished_tasks == 1:
                logger.debug("final hook is hanging")

    # stops execution of all daemons
    def free(self):
        self.running = False
        with self.mutex:
            self.queue.clear()


config = get_config()
handler = EventQueue(config.num_worker_threads)



class KubernetesCluster:
    """Kubernetes Cluster"""

    name = "Kubernetes Cluster"


class KubectlClient:
    """The kubectl client binary is used by the user to interact with the cluster"""

    name = "Kubectl Client"


class Kubelet(KubernetesCluster):
    """The kubelet is the primary "node agent" that runs on each node"""

    name = "Kubelet"


class AWS(KubernetesCluster):
    """AWS Cluster"""

    name = "AWS"


class Azure(KubernetesCluster):
    """Azure Cluster"""

    name = "Azure"



import time

from kube_hunter.conf import Config, set_config
from kube_hunter.core.types import Hunter
from kube_hunter.core.events.types import Event, Service
from kube_hunter.core.events.event_handler import handler

counter = 0
first_run = True

set_config(Config())


class OnceOnlyEvent(Service, Event):
    def __init__(self):
        Service.__init__(self, "Test Once Service")


class RegularEvent(Service, Event):
    def __init__(self):
        Service.__init__(self, "Test Service")


class AnotherRegularEvent(Service, Event):
    def __init__(self):
        Service.__init__(self, "Test Service (another)")


class DifferentRegularEvent(Service, Event):
    def __init__(self):
        Service.__init__(self, "Test Service (different)")


@handler.subscribe_once(OnceOnlyEvent)
class OnceHunter(Hunter):
    def __init__(self, event):
        global counter
        counter += 1


@handler.subscribe(RegularEvent)
class RegularHunter(Hunter):
    def __init__(self, event):
        global counter
        counter += 1


@handler.subscribe_many([DifferentRegularEvent, AnotherRegularEvent])
class SmartHunter(Hunter):
    def __init__(self, events):
        global counter, first_run
        counter += 1

        # we add an attribute on the second scan.
        # here we test that we get the latest event
        different_event = events.get_by_class(DifferentRegularEvent)
        if first_run:
            first_run = False
            assert not different_event.new_value
        else:
            assert different_event.new_value


@handler.subscribe_many([DifferentRegularEvent, AnotherRegularEvent])
class SmartHunter2(Hunter):
    def __init__(self, events):
        global counter
        counter += 1

        # check if we can access the events
        assert events.get_by_class(DifferentRegularEvent).__class__ == DifferentRegularEvent
        assert events.get_by_class(AnotherRegularEvent).__class__ == AnotherRegularEvent


def test_subscribe_mechanism():
    global counter
    counter = 0

    # first test normal subscribe and publish works
    handler.publish_event(RegularEvent())
    handler.publish_event(RegularEvent())
    handler.publish_event(RegularEvent())

    time.sleep(0.02)
    assert counter == 3


def test_subscribe_once_mechanism():
    global counter
    counter = 0

    # testing the multiple subscription mechanism
    handler.publish_event(OnceOnlyEvent())

    time.sleep(0.02)
    assert counter == 1
    counter = 0

    handler.publish_event(OnceOnlyEvent())
    handler.publish_event(OnceOnlyEvent())
    handler.publish_event(OnceOnlyEvent())
    time.sleep(0.02)

    assert counter == 0


def test_subscribe_many_mechanism():
    global counter
    counter = 0

    # testing the multiple subscription mechanism
    handler.publish_event(DifferentRegularEvent())
    handler.publish_event(DifferentRegularEvent())
    handler.publish_event(DifferentRegularEvent())
    handler.publish_event(DifferentRegularEvent())
    handler.publish_event(DifferentRegularEvent())
    handler.publish_event(AnotherRegularEvent())

    time.sleep(0.02)
    # We expect SmartHunter and SmartHunter2 to be executed once. hence the counter should be 2
    assert counter == 2
    counter = 0

    # Test using most recent event
    newer_version_event = DifferentRegularEvent()
    newer_version_event.new_value = True
    handler.publish_event(newer_version_event)

    assert counter == 2



import os
import json
import logging
import requests

from kube_hunter.conf import get_config
from kube_hunter.modules.hunting.kubelet import ExposedPodsHandler, SecureKubeletPortHunter
from kube_hunter.core.events.event_handler import handler
from kube_hunter.core.events.types import Event, Vulnerability
from kube_hunter.core.types import Hunter, ActiveHunter, MountServicePrincipalTechnique, Azure

logger = logging.getLogger(__name__)


class AzureSpnExposure(Vulnerability, Event):
    """The SPN is exposed, potentially allowing an attacker to gain access to the Azure subscription"""

    def __init__(self, container, evidence=""):
        Vulnerability.__init__(
            self,
            Azure,
            "Azure SPN Exposure",
            category=MountServicePrincipalTechnique,
            vid="KHV004",
        )
        self.container = container
        self.evidence = evidence


@handler.subscribe(ExposedPodsHandler, predicate=lambda x: x.cloud_type == "Azure")
class AzureSpnHunter(Hunter):
    """AKS Hunting
    Hunting Azure cluster deployments using specific known configurations
    """

    def __init__(self, event):
        self.event = event
        self.base_url = f"https://{self.event.host}:{self.event.port}"

    # getting a container that has access to the azure.json file
    def get_key_container(self):
        logger.debug("Trying to find container with access to azure.json file")

        # pods are saved in the previous event object
        pods_data = self.event.pods

        suspicious_volume_names = []
        for pod_data in pods_data:
            for volume in pod_data["spec"].get("volumes", []):
                if volume.get("hostPath"):
                    path = volume["hostPath"]["path"]
                    if "/etc/kubernetes/azure.json".startswith(path):
                        suspicious_volume_names.append(volume["name"])
            for container in pod_data["spec"]["containers"]:
                for mount in container.get("volumeMounts", []):
                    if mount["name"] in suspicious_volume_names:
                        return {
                            "name": container["name"],
                            "pod": pod_data["metadata"]["name"],
                            "namespace": pod_data["metadata"]["namespace"],
                            "mount": mount,
                        }

    def execute(self):
        container = self.get_key_container()
        if container:
            evidence = f"pod: {container['pod']}, namespace: {container['namespace']}"
            self.publish_event(AzureSpnExposure(container=container, evidence=evidence))


@handler.subscribe(AzureSpnExposure)
class ProveAzureSpnExposure(ActiveHunter):
    """Azure SPN Hunter
    Gets the azure subscription file on the host by executing inside a container
    """

    def __init__(self, event):
        self.event = event
        self.base_url = f"https://{self.event.host}:{self.event.port}"

    def test_run_capability(self):
        """
        Uses SecureKubeletPortHunter to test the /run handler
        TODO: when multiple event subscription is implemented, use this here to make sure /run is accessible
        """
        debug_handlers = SecureKubeletPortHunter.DebugHandlers(path=self.base_url, session=self.event.session, pod=None)
        return debug_handlers.test_run_container()

    def run(self, command, container):
        config = get_config()
        run_url = f"{self.base_url}/run/{container['namespace']}/{container['pod']}/{container['name']}"
        return self.event.session.post(run_url, verify=False, params={"cmd": command}, timeout=config.network_timeout)

    def get_full_path_to_azure_file(self):
        """
        Returns a full path to /etc/kubernetes/azure.json
        Taking into consideration the difference folder of the mount inside the container.
        TODO: implement the edge case where the mount is to parent /etc folder.
        """
        azure_file_path = self.event.container["mount"]["mountPath"]

        # taking care of cases where a subPath is added to map the specific file
        if not azure_file_path.endswith("azure.json"):
            azure_file_path = os.path.join(azure_file_path, "azure.json")

        return azure_file_path

    def execute(self):
        if not self.test_run_capability():
            logger.debug("Not proving AzureSpnExposure because /run debug handler is disabled")
            return

        try:
            azure_file_path = self.get_full_path_to_azure_file()
            logger.debug(f"trying to access the azure.json at the resolved path: {azure_file_path}")
            subscription = self.run(f"cat {azure_file_path}", container=self.event.container).json()
        except requests.Timeout:
            logger.debug("failed to run command in container", exc_info=True)
        except json.decoder.JSONDecodeError:
            logger.warning("failed to parse SPN")
        else:
            if "subscriptionId" in subscription:
                self.event.subscriptionId = subscription["subscriptionId"]
                self.event.aadClientId = subscription["aadClientId"]
                self.event.aadClientSecret = subscription["aadClientSecret"]
                self.event.tenantId = subscription["tenantId"]
                self.event.evidence = f"subscription: {self.event.subscriptionId}"



from kube_hunter.conf import Config, set_config

from kube_hunter.modules.discovery.kubernetes_client import list_all_k8s_cluster_nodes
from unittest.mock import MagicMock, patch

set_config(Config())


def test_client_yields_ips():
    client = MagicMock()
    response = MagicMock()
    client.list_node.return_value = response
    response.items = [MagicMock(), MagicMock()]
    response.items[0].status.addresses = [MagicMock(), MagicMock()]
    response.items[0].status.addresses[0].address = "127.0.0.1"
    response.items[0].status.addresses[1].address = "127.0.0.2"
    response.items[1].status.addresses = [MagicMock()]
    response.items[1].status.addresses[0].address = "127.0.0.3"

    with patch("kubernetes.config.load_incluster_config") as m:
        output = list(list_all_k8s_cluster_nodes(client=client))
        m.assert_called_once()

    assert output == ["127.0.0.1", "127.0.0.2", "127.0.0.3"]


def test_client_uses_kubeconfig():
    with patch("kubernetes.config.load_kube_config") as m:
        list(list_all_k8s_cluster_nodes(kube_config="/location", client=MagicMock()))
        m.assert_called_once_with(config_file="/location")



# flake8: noqa: E402
from kube_hunter.modules.discovery.hosts import (
    FromPodHostDiscovery,
    RunningAsPodEvent,
    HostScanEvent,
    HostDiscoveryHelpers,
)
from kube_hunter.core.types import Hunter
from kube_hunter.core.events.event_handler import handler
import json
import requests_mock
import pytest

from netaddr import IPNetwork, IPAddress
from typing import List
from kube_hunter.conf import Config, get_config, set_config

set_config(Config())


class TestFromPodHostDiscovery:
    @staticmethod
    def _make_azure_response(*subnets: List[tuple]) -> str:
        return json.dumps(
            {
                "network": {
                    "interface": [
                        {"ipv4": {"subnet": [{"address": address, "prefix": prefix} for address, prefix in subnets]}}
                    ]
                }
            }
        )

    @staticmethod
    def _make_aws_response(*data: List[str]) -> str:
        return "\n".join(data)

    def test_is_azure_pod_request_fail(self):
        f = FromPodHostDiscovery(RunningAsPodEvent())

        with requests_mock.Mocker() as m:
            m.get("http://169.254.169.254/metadata/instance?api-version=2017-08-01", status_code=404)
            result = f.is_azure_pod()

        assert not result

    def test_is_azure_pod_success(self):
        f = FromPodHostDiscovery(RunningAsPodEvent())

        with requests_mock.Mocker() as m:
            m.get(
                "http://169.254.169.254/metadata/instance?api-version=2017-08-01",
                text=TestFromPodHostDiscovery._make_azure_response(("3.4.5.6", "255.255.255.252")),
            )
            result = f.is_azure_pod()

        assert result

    def test_is_aws_pod_v1_request_fail(self):
        f = FromPodHostDiscovery(RunningAsPodEvent())

        with requests_mock.Mocker() as m:
            m.get("http://169.254.169.254/latest/meta-data/", status_code=404)
            result = f.is_aws_pod_v1()

        assert not result

    def test_is_aws_pod_v1_success(self):
        f = FromPodHostDiscovery(RunningAsPodEvent())

        with requests_mock.Mocker() as m:
            m.get(
                "http://169.254.169.254/latest/meta-data/",
                text=TestFromPodHostDiscovery._make_aws_response(
                    "\n".join(
                        (
                            "ami-id",
                            "ami-launch-index",
                            "ami-manifest-path",
                            "block-device-mapping/",
                            "events/",
                            "hostname",
                            "iam/",
                            "instance-action",
                            "instance-id",
                            "instance-type",
                            "local-hostname",
                            "local-ipv4",
                            "mac",
                            "metrics/",
                            "network/",
                            "placement/",
                            "profile",
                            "public-hostname",
                            "public-ipv4",
                            "public-keys/",
                            "reservation-id",
                            "security-groups",
                            "services/",
                        )
                    ),
                ),
            )
            result = f.is_aws_pod_v1()

        assert result

    def test_is_aws_pod_v2_request_fail(self):
        f = FromPodHostDiscovery(RunningAsPodEvent())

        with requests_mock.Mocker() as m:
            m.put(
                "http://169.254.169.254/latest/api/token/",
                headers={"X-aws-ec2-metatadata-token-ttl-seconds": "21600"},
                status_code=404,
            )
            m.get(
                "http://169.254.169.254/latest/meta-data/",
                headers={"X-aws-ec2-metatadata-token": "token"},
                status_code=404,
            )
            result = f.is_aws_pod_v2()

        assert not result

    def test_is_aws_pod_v2_success(self):
        f = FromPodHostDiscovery(RunningAsPodEvent())

        with requests_mock.Mocker() as m:
            m.put(
                "http://169.254.169.254/latest/api/token/",
                headers={"X-aws-ec2-metatadata-token-ttl-seconds": "21600"},
                text=TestFromPodHostDiscovery._make_aws_response("token"),
            )
            m.get(
                "http://169.254.169.254/latest/meta-data/",
                headers={"X-aws-ec2-metatadata-token": "token"},
                text=TestFromPodHostDiscovery._make_aws_response(
                    "\n".join(
                        (
                            "ami-id",
                            "ami-launch-index",
                            "ami-manifest-path",
                            "block-device-mapping/",
                            "events/",
                            "hostname",
                            "iam/",
                            "instance-action",
                            "instance-id",
                            "instance-type",
                            "local-hostname",
                            "local-ipv4",
                            "mac",
                            "metrics/",
                            "network/",
                            "placement/",
                            "profile",
                            "public-hostname",
                            "public-ipv4",
                            "public-keys/",
                            "reservation-id",
                            "security-groups",
                            "services/",
                        )
                    ),
                ),
            )
            result = f.is_aws_pod_v2()

        assert result

    def test_execute_scan_cidr(self):
        set_config(Config(cidr="1.2.3.4/30"))
        f = FromPodHostDiscovery(RunningAsPodEvent())
        f.execute()

    def test_execute_scan_remote(self):
        set_config(Config(remote="1.2.3.4"))
        f = FromPodHostDiscovery(RunningAsPodEvent())
        f.execute()


@handler.subscribe(HostScanEvent)
class HunterTestHostDiscovery(Hunter):
    """TestHostDiscovery
    In this set of tests we should only trigger HostScanEvent when remote or cidr are set
    """

    def __init__(self, event):
        config = get_config()
        assert config.remote is not None or config.cidr is not None
        assert config.remote == "1.2.3.4" or config.cidr == "1.2.3.4/30"


class TestDiscoveryUtils:
    @staticmethod
    def test_generate_hosts_valid_cidr():
        test_cidr = "192.168.0.0/24"
        expected = set(IPNetwork(test_cidr))

        actual = set(HostDiscoveryHelpers.generate_hosts([test_cidr]))

        assert actual == expected

    @staticmethod
    def test_generate_hosts_valid_ignore():
        remove = IPAddress("192.168.1.8")
        scan = "192.168.1.0/24"
        expected = {ip for ip in IPNetwork(scan) if ip != remove}

        actual = set(HostDiscoveryHelpers.generate_hosts([scan, f"!{str(remove)}"]))

        assert actual == expected

    @staticmethod
    def test_generate_hosts_invalid_cidr():
        with pytest.raises(ValueError):
            list(HostDiscoveryHelpers.generate_hosts(["192..2.3/24"]))

    @staticmethod
    def test_generate_hosts_invalid_ignore():
        with pytest.raises(ValueError):
            list(HostDiscoveryHelpers.generate_hosts(["192.168.1.8", "!29.2..1/24"]))



class HunterBase:
    publishedVulnerabilities = 0

    @staticmethod
    def parse_docs(docs):
        """returns tuple of (name, docs)"""
        if not docs:
            return __name__, "<no documentation>"
        docs = docs.strip().split("\n")
        for i, line in enumerate(docs):
            docs[i] = line.strip()
        return docs[0], " ".join(docs[1:]) if len(docs[1:]) else "<no documentation>"

    @classmethod
    def get_name(cls):
        name, _ = cls.parse_docs(cls.__doc__)
        return name

    def publish_event(self, event):
        # Import here to avoid circular import from events package.
        # imports are cached in python so this should not affect runtime
        from ..events.event_handler import handler  # noqa

        handler.publish_event(event, caller=self)


class ActiveHunter(HunterBase):
    pass


class Hunter(HunterBase):
    pass


class Discovery(HunterBase):
    pass



"""
Vulnerabilities are divided into 2 main categories.

MITRE Category
--------------
Vulnerability that correlates to a method in the official MITRE ATT&CK matrix for kubernetes

CVE Category
-------------
"General" category definition. The category is usually determined by the severity of the CVE
"""


class MITRECategory:
    @classmethod
    def get_name(cls):
        """
        Returns the full name of MITRE technique: <MITRE CATEGORY> // <MITRE TECHNIQUE>
        Should only be used on a direct technique class at the end of the MITRE inheritance chain.

        Example inheritance:
        MITRECategory -> InitialAccessCategory -> ExposedSensitiveInterfacesTechnique
        """
        inheritance_chain = cls.__mro__
        if len(inheritance_chain) >= 4:
            # -3 == index of mitreCategory class. (object class is first)
            mitre_category_class = inheritance_chain[-3]
        return f"{mitre_category_class.name} // {cls.name}"


class CVECategory:
    @classmethod
    def get_name(cls):
        """
        Returns the full name of the category: CVE // <CVE Category name>
        """
        return f"CVE // {cls.name}"


"""
MITRE ATT&CK Technique Categories
"""


class InitialAccessCategory(MITRECategory):
    name = "Initial Access"


class ExecutionCategory(MITRECategory):
    name = "Execution"


class PersistenceCategory(MITRECategory):
    name = "Persistence"


class PrivilegeEscalationCategory(MITRECategory):
    name = "Privilege Escalation"


class DefenseEvasionCategory(MITRECategory):
    name = "Defense Evasion"


class CredentialAccessCategory(MITRECategory):
    name = "Credential Access"


class DiscoveryCategory(MITRECategory):
    name = "Discovery"


class LateralMovementCategory(MITRECategory):
    name = "Lateral Movement"


class CollectionCategory(MITRECategory):
    name = "Collection"


class ImpactCategory(MITRECategory):
    name = "Impact"


"""
MITRE ATT&CK Techniques
"""


class GeneralSensitiveInformationTechnique(InitialAccessCategory):
    name = "General Sensitive Information"


class ExposedSensitiveInterfacesTechnique(InitialAccessCategory):
    name = "Exposed sensitive interfaces"


class MountServicePrincipalTechnique(CredentialAccessCategory):
    name = "Mount service principal"


class ListK8sSecretsTechnique(CredentialAccessCategory):
    name = "List K8S secrets"


class AccessContainerServiceAccountTechnique(CredentialAccessCategory):
    name = "Access container service account"


class AccessK8sApiServerTechnique(DiscoveryCategory):
    name = "Access the K8S API Server"


class AccessKubeletAPITechnique(DiscoveryCategory):
    name = "Access Kubelet API"


class AccessK8sDashboardTechnique(DiscoveryCategory):
    name = "Access Kubernetes Dashboard"


class InstanceMetadataApiTechnique(DiscoveryCategory):
    name = "Instance Metadata API"


class ExecIntoContainerTechnique(ExecutionCategory):
    name = "Exec into container"


class SidecarInjectionTechnique(ExecutionCategory):
    name = "Sidecar injection"


class NewContainerTechnique(ExecutionCategory):
    name = "New container"


class GeneralPersistenceTechnique(PersistenceCategory):
    name = "General Peristence"


class HostPathMountPrivilegeEscalationTechnique(PrivilegeEscalationCategory):
    name = "hostPath mount"


class PrivilegedContainerTechnique(PrivilegeEscalationCategory):
    name = "Privileged container"


class ClusterAdminBindingTechnique(PrivilegeEscalationCategory):
    name = "Cluser-admin binding"


class ARPPoisoningTechnique(LateralMovementCategory):
    name = "ARP poisoning and IP spoofing"


class CoreDNSPoisoningTechnique(LateralMovementCategory):
    name = "CoreDNS poisoning"


class DataDestructionTechnique(ImpactCategory):
    name = "Data Destruction"


class GeneralDefenseEvasionTechnique(DefenseEvasionCategory):
    name = "General Defense Evasion"


class ConnectFromProxyServerTechnique(DefenseEvasionCategory):
    name = "Connect from Proxy server"


"""
CVE Categories
"""


class CVERemoteCodeExecutionCategory(CVECategory):
    name = "Remote Code Execution (CVE)"


class CVEPrivilegeEscalationCategory(CVECategory):
    name = "Privilege Escalation (CVE)"


class CVEDenialOfServiceTechnique(CVECategory):
    name = "Denial Of Service (CVE)"



import logging
import kubernetes


def list_all_k8s_cluster_nodes(kube_config=None, client=None):
    logger = logging.getLogger(__name__)
    try:
        if kube_config:
            logger.debug("Attempting to use kubeconfig file: %s", kube_config)
            kubernetes.config.load_kube_config(config_file=kube_config)
        else:
            logger.debug("Attempting to use in cluster Kubernetes config")
            kubernetes.config.load_incluster_config()
    except kubernetes.config.config_exception.ConfigException as ex:
        logger.debug(f"Failed to initiate Kubernetes client: {ex}")
        return

    try:
        if client is None:
            client = kubernetes.client.CoreV1Api()
        ret = client.list_node(watch=False)
        logger.info("Listed %d nodes in the cluster" % len(ret.items))
        for item in ret.items:
            for addr in item.status.addresses:
                yield addr.address
    except Exception as ex:
        logger.debug(f"Failed to list nodes from Kubernetes: {ex}")
